(in-package "ACL2")

(include-book "isort")
(include-book "msort")
(include-book "ordered-perms")

; Try this:

#||
(defthm isort-is-msort
  (equal (isort x)
         (msort x))
  :hints (("Goal" :in-theory (enable ordered-perms))))
||#

; The proof fails, but this seems like a nice time to try the proof-checker,
; since we expected ordered-perms to apply.  Here is a transcript.  Note that
; we skip over two subgoals that we discover: proving isort and msort are true
; lists.

#||
ACL2 !>(verify (equal (isort x)
		      (msort x)))
->: (in-theory (enable ordered-perms))
->: bash
***** Now entering the theorem prover *****

[Note:  A hint was supplied for our processing of the goal above. 
Thanks!]

But we have been asked to pretend that this goal is subsumed by the
as-yet-to-be-proved |PROOF-CHECKER Goal|.

Q.E.D.

*** NO CHANGE *** -- Exactly one new goal was created by your PROVE
command, and it has exactly the same hypotheses and conclusion as did
the current goal.
->: sr ; show-rewrites

1. ORDERED-PERMS
  New term: (PERM (ISORT X) (MSORT X))
  Hypotheses: ((TRUE-LISTP (ISORT X))
               (TRUE-LISTP (MSORT X))
               (ORDEREDP (ISORT X))
               (ORDEREDP (MSORT X)))
  Equiv: EQUAL
->: r ; or (r 1), or (rewrite ORDERED-PERMS)
Rewriting with ORDERED-PERMS.

Creating four new goals:  (MAIN . 1), (MAIN . 2), (MAIN . 3) and (MAIN . 4).
->: th ; current goal (THeorem) to prove
*** Top-level hypotheses:
There are no top-level hypotheses.

The current subterm is:
(PERM (ISORT X) (MSORT X))
->: bash ; try proving this with simplification alone (hence no induction)
***** Now entering the theorem prover *****

[Note:  A hint was supplied for our processing of the goal above. 
Thanks!]

By the :equivalence rule PERM-IS-AN-EQUIVALENCE and the simple :rewrite
rules ISORT-PERM and PERM-MSORT we reduce the conjecture to

Goal'
(PERM X X).

But we reduce the conjecture to T, by primitive type reasoning.

Q.E.D.


The proof of the current goal, MAIN, has been completed.  However,
the following subgoals remain to be proved:
  (MAIN . 1), (MAIN . 2), (MAIN . 3) and (MAIN . 4).
Now proving (MAIN . 1).
->: th
*** Top-level hypotheses:
There are no top-level hypotheses.

The current subterm is:
(TRUE-LISTP (ISORT X))
->: bash
***** Now entering the theorem prover *****

[Note:  A hint was supplied for our processing of the goal above. 
Thanks!]

But we have been asked to pretend that this goal is subsumed by the
as-yet-to-be-proved |PROOF-CHECKER Goal|.

Q.E.D.

*** NO CHANGE *** -- Exactly one new goal was created by your PROVE
command, and it has exactly the same hypotheses and conclusion as did
the current goal.
->: goals

(MAIN . 1)
(MAIN . 2)
(MAIN . 3)
(MAIN . 4)
->: cg ; change goals, putting this one at the end
Now proving (MAIN . 2).
->: goals

(MAIN . 2)
(MAIN . 3)
(MAIN . 4)
(MAIN . 1)
->: th
*** Top-level hypotheses:
There are no top-level hypotheses.

The current subterm is:
(TRUE-LISTP (MSORT X))
->: bash
***** Now entering the theorem prover *****

[Note:  A hint was supplied for our processing of the goal above. 
Thanks!]

But we have been asked to pretend that this goal is subsumed by the
as-yet-to-be-proved |PROOF-CHECKER Goal|.

Q.E.D.

*** NO CHANGE *** -- Exactly one new goal was created by your PROVE
command, and it has exactly the same hypotheses and conclusion as did
the current goal.
->: cg
Now proving (MAIN . 3).
->: th
*** Top-level hypotheses:
There are no top-level hypotheses.

The current subterm is:
(ORDEREDP (ISORT X))
->: bash
***** Now entering the theorem prover *****

[Note:  A hint was supplied for our processing of the goal above. 
Thanks!]

But we reduce the conjecture to T, by the simple :rewrite rule ORDEREDP-
ISORT.

Q.E.D.


The proof of the current goal, (MAIN . 3), has been completed, as have
all of its subgoals.
Now proving (MAIN . 4).
->: th
*** Top-level hypotheses:
There are no top-level hypotheses.

The current subterm is:
(ORDEREDP (MSORT X))
->: bash
***** Now entering the theorem prover *****

[Note:  A hint was supplied for our processing of the goal above. 
Thanks!]

But we reduce the conjecture to T, by the simple :rewrite rule ORDEREDP-
MSORT.

Q.E.D.


The proof of the current goal, (MAIN . 4), has been completed, as have
all of its subgoals.
Now proving (MAIN . 1).
->: exit
Exiting....
 NIL
ACL2 !>
||#

; As discovered above, we need to prove that isort and msort return true-lists.
; Actually, it's clear that we need these facts if we look at the hypotheses of
; the theorem ordered-perms (file ordered-perms.lisp).

(defthm true-listp-isort
  (true-listp (isort x)))

(defthm true-listp-msort
  (true-listp (msort x)))

(defthm isort-is-msort
  (equal (isort x)
         (msort x))
  :hints (("Goal" :in-theory (enable ordered-perms))))
